﻿using System.Collections.Generic;

namespace TestProject
{
    /// <summary>
    /// 스레드 안전 우선 순위 큐
    /// </summary>
    /// <typeparam name="TItemKey">항목 키 타입</typeparam>
    /// <typeparam name="TIem">항목 타입</typeparam>
    public class ThreadSafePriorityQueue<TItemKey, TItem> : Dictionary<int, ThreadSafeSortedListQueue<TItemKey, TItem>> where TItem : IKeyMessage<TItemKey>, IPriorityMessage
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region Field

        /// <summary>
        /// 동기 객체
        /// </summary>
        protected object syncObject;

        /// <summary>
        /// 우선 순위 카운트
        /// </summary>
        protected int priorityCount;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 우선 순위 카운트 - PriorityCount

        /// <summary>
        /// 우선 순위 카운트
        /// </summary>
        public int PriorityCount
        {
            get
            {
                return this.priorityCount;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - ThreadSafePriorityQueue(priorityCount)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="priorityCount">우선 순위 카운트</param>
        public ThreadSafePriorityQueue(int priorityCount)
        {
            this.syncObject    = new object();
            this.priorityCount = priorityCount;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 항목 넣기 - EnqueueItem(item)

        /// <summary>
        /// 항목 넣기
        /// </summary>
        /// <param name="item">항목</param>
        public void EnqueueItem(TItem item)
        {
            lock(this.syncObject)
            {
                int priority = item.Priority;

                if(priority < 0)
                {
                    priority = 0;
                }

                if(priority >= this.priorityCount)
                {
                    priority = this.priorityCount - 1;
                }

                ThreadSafeSortedListQueue<TItemKey, TItem> queue;

                if(base.ContainsKey(priority))
                {
                    queue = this[priority];
                }
                else
                {
                    queue = new ThreadSafeSortedListQueue<TItemKey, TItem>();

                    base.Add(priority, queue);
                }

                queue.EnqueueItem(item.Key, item);
            }
        }

        #endregion
        #region 항목 꺼내기 - DequeueItem()

        /// <summary>
        /// 항목 꺼내기
        /// </summary>
        /// <returns>항목</returns>
        public TItem DequeueItem()
        {
            lock(this.syncObject)
            {
                for(int i = this.priorityCount - 1; i >= 0; i--)
                {
                    if(ContainsKey(i))
                    {
                        ThreadSafeSortedListQueue<TItemKey, TItem> queue = this[i];

                        if(queue.Count > 0)
                        {
                            TItem item = queue.DequeueItem();

                            return item;
                        }
                    }
                }

                return default(TItem);
            }
        }

        #endregion
        #region 항목 카운트 구하기 - GetItemCount()

        /// <summary>
        /// 항목 카운트 구하기
        /// </summary>
        /// <returns>항목 카운트</returns>
        public int GetItemCount()
        {
            lock(this.syncObject)
            {
                int count = 0;

                foreach(KeyValuePair<int, ThreadSafeSortedListQueue<TItemKey, TItem>> keyValuePair in this)
                {
                    count += keyValuePair.Value.Count;
                }

                return count;
            }
        }

        #endregion
        #region 데이터 지우기 - ClearData()

        /// <summary>
        /// 데이터 지우기
        /// </summary>
        public void ClearData()
        {
            lock(this.syncObject)
            {
                foreach(KeyValuePair<int, ThreadSafeSortedListQueue<TItemKey, TItem>> keyValuePair in this)
                {
                    ThreadSafeSortedListQueue<TItemKey, TItem> queue = keyValuePair.Value;

                    queue.ClearData();
                }

                Clear();
            }
        }

        #endregion
    }
}